from functools import cached_property
from itertools import chain

from dictlib import dug
from django.db.models.sql.compiler import (
    SQLCompiler as BaseSQLCompiler,
)
from django.db.models.sql.compiler import (
    SQLInsertCompiler as BaseSQLInsertCompiler,
)
from django.db.models.sql.compiler import (
    cursor_iter,
)
from django.db.models.sql.constants import (
    CURSOR,
    GET_ITERATOR_CHUNK_SIZE,
    MULTI,
    NO_RESULTS,
    SINGLE,
)
from pymongo import InsertOne, UpdateOne

from django_mongodb.query import MongoOrdering, MongoSelect, MongoWhereNode


class SQLCompiler(BaseSQLCompiler):
    def __init__(self, query, connection, using, elide_empty=True):
        super().__init__(query, connection, using, elide_empty)
        self.extr = None

    def build_mongo_filter(self, filter_expr):
        referenced_tables = set()
        for key, item in self.query.alias_refcount.items():
            if item < 1:
                continue
            else:
                referenced_tables.add(self.query.alias_map[key].table_name)

        if len(referenced_tables) > 1:
            raise NotImplementedError("Multi-table joins are not implemented yet.")

        return MongoWhereNode(filter_expr, self.mongo_meta)

    def get_distinct_clause(self):
        return [
            {
                "$group": {
                    "_id": {
                        col.target.attname: f"${col.target.column}" for col, _, alias in self.select
                    },
                },
            },
            {"$replaceRoot": {"newRoot": "$_id"}},
        ]

    def as_operation(self, with_limits=True, with_col_aliases=False):  # noqa: C901
        combinator = self.query.combinator
        extra_select, order_by, group_by = self.pre_sql_setup(
            with_col_aliases=with_col_aliases or bool(combinator),
        )
        if combinator or extra_select or group_by or with_col_aliases:
            raise NotImplementedError

        # Is a LIMIT/OFFSET clause needed?
        with_limit_offset = with_limits and self.query.is_sliced
        if self.query.select_for_update:
            raise NotImplementedError

        pipeline = []
        mongo_where = self.build_mongo_filter(self.query.where)
        build_search_pipeline = (
            (hasattr(self.query, "prefer_search") and self.query.prefer_search)
            or mongo_where.requires_search()
        ) and not self.query.distinct  # search not supported / efficient for distinct queries

        self._extend_with_stage(pipeline, "prepend")

        has_attname_as_key = False
        if mongo_where and build_search_pipeline:
            search = mongo_where.get_mongo_search(self, self.connection)
            order = MongoOrdering(self.query).get_mongo_order()
            if search:
                pipeline.append({"$search": search})
                if self.query.order_by:
                    search["sort"] = {**order}
            # we need to recheck fields, which did not have a search index
            if extra_match := mongo_where.get_mongo_query(self, self.connection, is_search=True):
                pipeline.append({"$match": extra_match})
        elif mongo_where:
            pipeline.append(
                {"$match": mongo_where.get_mongo_query(self, self.connection, is_search=False)}
            )

        self._extend_with_stage(pipeline, "pre-sort")

        if self.query.distinct:
            has_attname_as_key = True
            pipeline.extend(self.get_distinct_clause())

        if self.query.order_by and not build_search_pipeline:
            order = MongoOrdering(self.query).get_mongo_order(attname_as_key=has_attname_as_key)
            pipeline.append({"$sort": order})

        if with_limit_offset and self.query.low_mark:
            pipeline.append({"$skip": self.query.low_mark})

        if with_limit_offset and self.query.high_mark:
            pipeline.append({"$limit": self.query.high_mark - self.query.low_mark})

        self._extend_with_stage(pipeline, "append")

        if (select_cols := self.select + extra_select) and not has_attname_as_key:
            select_pipeline = MongoSelect(select_cols, self.mongo_meta).get_mongo()
            pipeline.extend(select_pipeline)

        return {
            "collection": self.query.model._meta.db_table,
            "op": "aggregate",
            "pipeline": [
                *pipeline,
            ],
        }

    def _extend_with_stage(self, pipeline, position):
        if not hasattr(self.query, "aggregation_stages"):
            return
        if self.query.aggregation_stages and any(
            stages := [stage for pos, stage in self.query.aggregation_stages if pos == position]
        ):
            pipeline.extend(stages)

    @cached_property
    def mongo_meta(self):
        if hasattr(self.query.model, "MongoMeta"):
            _meta = self.query.model.MongoMeta
            return {
                "search_fields": {} if not hasattr(_meta, "search_fields") else _meta.search_fields
            }
        else:
            return {"search_fields": {}}

    def execute_sql(
        self, result_type=MULTI, chunked_fetch=False, chunk_size=GET_ITERATOR_CHUNK_SIZE
    ):
        result_type = result_type or NO_RESULTS
        self.setup_query()

        # Join handling, currently only pseudo-joins on same collection
        # for same collection inheritance
        if self.query.extra_tables:
            raise NotImplementedError("Can't do sub-queries with multiple tables yet.")

        cursor = self.connection.cursor()
        try:
            cursor.execute(self.as_operation())
        except Exception:
            cursor.close()
            raise

        if result_type == CURSOR:
            return cursor
        if result_type == SINGLE:
            cols = list(self.select)
            result = cursor.fetchone()
            if result:
                return (result.get(alias or col.target.attname) for col, _, alias in cols)
            return result
        if result_type == NO_RESULTS:
            cursor.close()
            return

        result = cursor_iter(
            cursor,
            self.connection.features.empty_fetchmany_value,
            None,
            chunk_size,
        )
        return result

    def apply_converters(self, rows, converters):
        connection = self.connection
        converters = list(converters.items())
        for row in map(list, rows):
            for pos, (convs, expression) in converters:
                value = row[pos]
                for converter in convs:
                    # MongoDB returns JSON fields as native dict already
                    if expression.output_field.db_type(connection) == "json":
                        continue
                    value = converter(value, expression, connection)
                row[pos] = value
            yield row

    def results_iter(
        self,
        results=None,
        tuple_expected=False,
        chunked_fetch=False,
        chunk_size=GET_ITERATOR_CHUNK_SIZE,
    ):
        """Return an iterator over the results from executing this query."""
        if results is None:
            results = self.execute_sql(MULTI, chunked_fetch=chunked_fetch, chunk_size=chunk_size)
        fields = [s[0] for s in self.select[0 : self.col_count]]
        converters = self.get_converters(fields)
        rows = chain.from_iterable(results)
        _row_tuples = []
        cols = self.select[0 : self.col_count]
        for row in rows:
            _row_tuples.append(
                tuple(row.get(alias or col.target.attname) for col, _, alias in cols)
            )

        if converters:
            _row_tuples = self.apply_converters(_row_tuples, converters)
            if tuple_expected:
                _row_tuples = map(tuple, _row_tuples)
        for row in _row_tuples:
            yield row


class SQLDeleteCompiler(SQLCompiler):
    def as_operation(self, with_limits=True, with_col_aliases=False):
        opts = self.query.get_meta()
        filter = self.build_mongo_filter(self.query.where).get_mongo_query(self, self.connection)
        return {
            "collection": opts.db_table,
            "op": "delete_many",
            "filter": filter,
        }


class SQLInsertCompiler(SQLCompiler, BaseSQLInsertCompiler):
    compiler = "SQLInsertCompiler"

    def as_operation(self):
        opts = self.query.get_meta()
        fields = self.query.fields or [opts.pk]

        if len(self.query.objs) == 1 and self.returning_fields:
            return {
                "collection": opts.db_table,
                "op": "insert_one",
                "document": self._fields_to_doc(fields, self.query.objs[0]),
            }
        elif self.returning_fields:
            raise NotImplementedError(
                "Returning fields is not supported for bulk inserts, right now, though very possible"
            )
        elif self.query.fields:
            insert_statement = [
                InsertOne(self._fields_to_doc(fields, obj))
                if not obj.pk
                #  need to upsert if pk is given, because we support single collection inheritance
                else UpdateOne(
                    {opts.pk.column: obj.pk},
                    {
                        "$set": {
                            field.column: self.prepare_value(field, self.pre_save_val(field, obj))
                            for field in fields
                            if not field.primary_key
                        }
                    },
                    upsert=True,
                )
                for obj in self.query.objs
            ]
        else:
            # An empty object.
            insert_statement = [
                InsertOne({opts.pk.column: self.connection.ops.pk_default_value()})
                for _ in self.query.objs
            ]

        return {
            "collection": opts.db_table,
            "op": "bulk_write",
            "requests": insert_statement,
        }

    def _fields_to_doc(self, fields, obj):
        doc = {}
        for field in fields:
            dug(doc, field.column, self.prepare_value(field, self.pre_save_val(field, obj)))
        return doc

    def execute_sql(self, returning_fields=None):
        assert not (
            returning_fields
            and len(self.query.objs) != 1
            and not self.connection.features.can_return_rows_from_bulk_insert
        )
        opts = self.query.get_meta()
        self.returning_fields = returning_fields
        with self.connection.cursor() as cursor:
            cursor.execute(self.as_operation(), None)
            if not self.returning_fields:
                return []
            else:
                rows = [
                    (
                        self.connection.ops.last_insert_id(
                            cursor,
                            opts.db_table,
                            opts.pk.column,
                        ),
                    )
                ]
        return rows


class SQLUpdateCompiler(SQLCompiler):
    def as_operation(self):
        opts = self.query.get_meta()
        filter = self.build_mongo_filter(self.query.where).get_mongo_query(self, self.connection)
        update = {
            field[0].column: field[0].get_db_prep_save(field[2], self.connection)
            for field in self.query.values
        }
        return {
            "collection": opts.db_table,
            "op": "update_many",
            "filter": filter,
            "update": {"$set": update},
        }

    def execute_sql(self, result_type):
        """
        Execute the specified update. Return the number of rows affected by
        the primary update query. The "primary update query" is the first
        non-empty query that is executed. Row counts for any subsequent,
        related queries are not available.
        """
        cursor = self.connection.cursor()
        try:
            cursor.execute(self.as_operation())
            rows = cursor.rowcount if cursor else 0
            is_empty = cursor is None
        finally:
            if cursor:
                cursor.close()
        for query in self.query.get_related_updates():
            aux_rows = query.get_compiler(self.using).execute_sql(result_type)
            if is_empty and aux_rows:
                rows = aux_rows
                is_empty = False
        return rows



import abc
from abc import ABC, abstractmethod
from collections import OrderedDict

from django.contrib.postgres.search import SearchQuery, SearchVector, SearchVectorExact
from django.db.models import Count
from django.db.models.expressions import BaseExpression, Col, Expression, Value
from django.db.models.fields.related_lookups import RelatedExact, RelatedIn
from django.db.models.lookups import (
    Exact,
    GreaterThan,
    GreaterThanOrEqual,
    In,
    IntegerFieldExact,
    IntegerGreaterThan,
    IntegerGreaterThanOrEqual,
    IntegerLessThan,
    IntegerLessThanOrEqual,
    IsNull,
    LessThan,
    LessThanOrEqual,
    Lookup,
)
from django.db.models.sql import Query
from django.db.models.sql.where import NothingNode, WhereNode

from django_mongodb.expressions import RawMongoDBQuery


class RequiresSearchException(Exception):
    pass


class RequiresSearchIndex(Exception):
    pass


class Node(ABC):
    def __init__(self, node: Expression, mongo_meta):
        self.node = node
        self.mongo_meta = mongo_meta

    def requires_search(self) -> bool:
        return False

    @abc.abstractmethod
    def get_mongo_query(self, compiler, connection, requires_search=...) -> dict: ...

    @abc.abstractmethod
    def get_mongo_search(self, compiler, connection) -> dict: ...


class RawMongoQueryExpression(Node):
    def __init__(self, node: RawMongoDBQuery, mongo_meta):
        super().__init__(node, mongo_meta)
        self.node = node

    def get_mongo_query(self, compiler, connection, is_search=False) -> dict:
        return self.node.query

    def get_mongo_search(self, compiler, connection) -> dict:
        return {}


class MongoLookup(Node):
    """MongoDB Query Node"""

    filter_operator: str

    def __init__(self, node: Lookup, mongo_meta):
        super().__init__(node, mongo_meta)
        self.lhs = node.get_prep_lhs()
        self.rhs = node.get_prep_lookup()
        if hasattr(self.node, "rhs") and isinstance(self.node.rhs, BaseExpression):
            raise NotImplementedError(f"Subquery Expression not implemented: {str(self.node.rhs)}")

    def get_mongo_query(self, compiler, connection, is_search=False) -> dict:
        if self.lhs.target.attname in self.mongo_meta["search_fields"] and is_search:
            return {}
        else:
            return self._get_mongo_query(compiler, connection)

    def _get_mongo_query(self, compiler, connection, is_search=False) -> dict:
        if self.lhs.target.attname in self.mongo_meta["search_fields"] and is_search:
            return {}
        lhs = self.lhs.target
        rhs = self.rhs
        if is_search and self.mongo_meta["search_fields"].get(lhs.attname):
            return {}
        return {lhs.column: {self.filter_operator: rhs}}

    def get_mongo_search(self, compiler, connection) -> dict:
        if self.lhs.target.attname not in self.mongo_meta["search_fields"]:
            return {}
        if not self.rhs:
            return {}
        else:
            return self._get_mongo_search(compiler, connection)

    @abc.abstractmethod
    def _get_mongo_search(self, compiler, connection) -> dict: ...


class MongoExact(MongoLookup):
    """MongoDB Query Node for Exact"""

    filter_operator = "$eq"

    def get_search_types(self, attname):
        if attname in self.mongo_meta["search_fields"]:
            if "string" in self.mongo_meta["search_fields"][attname]:
                return "text", "query"
            else:
                return "equals", "value"

    def _get_mongo_search(self, compiler, connection) -> dict:
        if self.lhs.target.attname not in self.mongo_meta["search_fields"]:
            return {}
        query, value = self.get_search_types(self.lhs.target.attname)
        return {
            query: {
                "path": self.lhs.target.column,
                value: self.rhs,
            }
        }


class MongoIn(MongoLookup):
    """MongoDB Query Node for RelatedIn"""

    filter_operator = "$in"

    def _get_mongo_search(self, compiler, connection) -> dict:
        if self.lhs.target.attname not in self.mongo_meta["search_fields"]:
            return {}
        return {
            "in": {
                "path": self.lhs.target.column,
                "value": self.rhs,
            }
        }


class MongoRelatedIn(MongoIn):
    filter_operator = "$in"


class MongoEqualityComparison(MongoLookup):
    """MongoDB Query Node for LessThanOrEqual"""

    filter_operator: str

    def __init__(
        self,
        operator: LessThan
        | LessThanOrEqual
        | GreaterThan
        | GreaterThanOrEqual
        | IntegerLessThan
        | IntegerLessThanOrEqual
        | IntegerGreaterThan,
        mongo_meta,
    ):
        super().__init__(operator, mongo_meta)
        self.filter_operator = {
            IntegerLessThan: "$lt",
            LessThan: "$lt",
            IntegerLessThanOrEqual: "$lt",
            LessThanOrEqual: "$lte",
            IntegerGreaterThan: "$gt",
            GreaterThan: "$gt",
            IntegerGreaterThanOrEqual: "$gte",
            GreaterThanOrEqual: "$gte",
        }[type(operator)]

    def _get_mongo_search(self, compiler, connection) -> dict:
        return {
            "range": {
                "path": self.lhs.target.column,
                self.filter_operator[1:-1]: self.rhs,
            }
        }


class MongoIsNull(MongoLookup):
    def _get_mongo_query(self, compiler, connection, is_search=False) -> dict:
        return {self.lhs.target.column: None if self.rhs else {"$ne": None}}

    def _get_mongo_search(self, compiler, connection) -> dict:
        if self.lhs.target.attname not in self.mongo_meta["search_fields"]:
            return {}
        return {
            "exists": {
                "path": self.lhs.target.column,
                "value": self.rhs,
            }
        }


class SearchNode(Node):
    """MongoDB Search Query Base Node"""

    def __init__(self, node: Expression, mongo_meta):
        super().__init__(node, mongo_meta)

    def requires_search(self) -> bool:
        return True

    def get_mongo_query(self, compiler, connection, is_search=False) -> dict:
        if not is_search:
            raise RequiresSearchException("SearchNode requires application in search pipeline.")
        else:
            return {}

    def get_mongo_search(self, compiler, connection) -> dict:
        return self._get_mongo_search(compiler, connection)

    @abstractmethod
    def _get_mongo_search(self, compiler, connection) -> dict: ...


class MongoSearchLookup(SearchNode):
    """MongoDB Search Query Node for SearchVectorExact"""

    def __init__(self, exact: Lookup, mongo_meta):
        super().__init__(exact, mongo_meta)
        self.lhs: SearchVector = exact.lhs
        self.rhs: SearchQuery = exact.rhs


class MongoSearchVectorExact(MongoSearchLookup):
    """Maps search vector to basic MongoDB wildcard query"""

    def _get_mongo_search(self, compiler, connection) -> dict:
        rhs_expressions = self.lhs.get_source_expressions()
        lhs_expressions = self.rhs.get_source_expressions()
        attname_to_column = {
            expression.field.attname: expression.field.column for expression in rhs_expressions
        }
        query = [expression.value for expression in lhs_expressions]
        if set(attname_to_column.keys()) - set(self.mongo_meta["search_fields"].keys()):
            raise RequiresSearchIndex(
                "SearchVectorExact requires a search index for the fields used in the search."
            )
        # weight = lhs.weight
        # config = lhs.config
        auto_complete_columns = [
            attname_to_column[key]
            for key, value in self.mongo_meta["search_fields"].items()
            if "autocomplete" in value and key in attname_to_column
        ]
        query_columns = [
            attname_to_column[key]
            for key, value in self.mongo_meta["search_fields"].items()
            if "string" in value and key not in auto_complete_columns and key in attname_to_column
        ]

        search_query = dict()
        if auto_complete_columns and query_columns:
            search_query = {
                "compound": {
                    "should": [
                        {"wildcard": {"path": query_columns, "query": query}},
                        *[
                            {"autocomplete": {"path": column, "query": query}}
                            for column in auto_complete_columns
                        ],
                    ]
                }
            }
        elif auto_complete_columns:
            search_query = {"autocomplete": {"path": auto_complete_columns, "query": query}}
        elif query_columns:
            search_query = {"wildcard": {"path": query_columns, "query": query}}
        return search_query


class MongoNothingNode(Node):
    def get_mongo_query(self, compiler, connection, is_search=...) -> dict:
        return {"$expr": {"$eq": [True, False]}}

    def get_mongo_search(self, compiler, connection) -> dict:
        return {}


class MongoWhereNode:
    """MongoDB Query Node for WhereNode"""

    node_map = {
        NothingNode: MongoNothingNode,
        Exact: MongoExact,
        IntegerFieldExact: MongoExact,
        RelatedIn: MongoRelatedIn,
        RelatedExact: MongoExact,
        In: MongoIn,
        LessThan: MongoEqualityComparison,
        IntegerLessThan: MongoEqualityComparison,
        LessThanOrEqual: MongoEqualityComparison,
        IntegerLessThanOrEqual: MongoEqualityComparison,
        GreaterThan: MongoEqualityComparison,
        IntegerGreaterThan: MongoEqualityComparison,
        GreaterThanOrEqual: MongoEqualityComparison,
        IntegerGreaterThanOrEqual: MongoEqualityComparison,
        SearchVectorExact: MongoSearchVectorExact,
        RawMongoDBQuery: RawMongoQueryExpression,
        IsNull: MongoIsNull,
    }

    def __init__(self, where: WhereNode, mongo_meta):
        self.node = where
        self.connector = where.connector
        self.children: list[MongoWhereNode | Node] = []
        self.mongo_meta = mongo_meta
        self.negated = where.negated
        for child in self.node.children:
            if isinstance(child, WhereNode):
                self.children.append(MongoWhereNode(child, self.mongo_meta))
            elif isinstance(child, Exact) and isinstance(child.lhs, RawMongoDBQuery):
                self.children.append(RawMongoQueryExpression(child.lhs, self.mongo_meta))
            elif child.__class__ in self.node_map:
                self.children.append(self.node_map[child.__class__](child, self.mongo_meta))
            else:
                raise NotImplementedError(f"Node not implemented: {type(child)}")

    def __bool__(self):
        return len(self.children) > 0

    def requires_search(self) -> bool:
        return any(child.requires_search() for child in self.children)

    def get_mongo_query(self, compiler, connection, is_search=False) -> dict:
        child_queries = list(
            filter(
                bool,
                [
                    child.get_mongo_query(compiler, connection, is_search=is_search)
                    for child in self.children
                ],
            )
        )
        if len(child_queries) == 0:
            return {}
        if self.connector == "AND":
            return {"$and": child_queries} if not self.negated else {"$nor": child_queries}
        elif self.connector == "OR":
            return {"$or": child_queries} if not self.negated else {"$nor": child_queries}
        else:
            raise Exception(f"Unsupported connector: {self.connector}")

    def get_mongo_search(self, compiler, connection) -> dict:
        child_queries = list(
            filter(
                bool,
                [child.get_mongo_search(compiler, connection) for child in self.children],
            )
        )
        if len(child_queries) == 0:
            return {}
        elif self.connector == "AND":
            return {"compound": {"must": child_queries}}
        elif self.connector == "OR":
            return {"compound": {"should": child_queries}}
        else:
            raise Exception(f"Unsupported connector: {self.connector}")


class MongoColSelect:
    def __init__(self, col: Col, alias: str | None, mongo_meta):
        self.col = col
        self.mongo_meta = mongo_meta
        self.alias = alias

    def get_mongo(self):
        return {"$project": {(self.alias or self.col.target.attname): f"${self.col.target.column}"}}


class MongoValueSelect:
    def __init__(self, col: Value, alias: str | None, mongo_meta):
        self.col = col
        self.mongo_meta = mongo_meta
        self.alias = alias

    def get_mongo(self):
        return {"$project": {(self.alias): self.col.value}}


class MongoCountSelect:
    def __init__(self, col: Count, alias: str | None, mongo_meta):
        self.col = col
        self.mongo_meta = mongo_meta
        self.alias = alias

    def get_mongo(self):
        return {
            "$group": {"_id": None, "_count": {"$sum": 1}},
            "$project": {
                "_id": None,
                (self.alias or self.col.output_field.column): "$_count",
            },
        }


class MongoSelect:
    def __init__(self, _cols: list[tuple[Expression, tuple, str | None]], mongo_meta):
        self.mongo_meta = mongo_meta
        self.cols = []
        for col in _cols:
            [column, _, alias] = col
            match column:
                case Col():
                    self.cols.append(MongoColSelect(column, alias, mongo_meta))
                case Value():
                    self.cols.append(MongoValueSelect(column, alias, mongo_meta))
                case Count():
                    self.cols.append(MongoCountSelect(column, alias, mongo_meta))
                case SearchVector():
                    pass  # ignoring search vector in results
                case _:
                    raise NotImplementedError(f"Select expression not implemented: {col}")

    def get_mongo(self):
        pipeline_dict: dict[str, dict] = OrderedDict()
        pipeline_dict["$group"] = dict()
        pipeline_dict["$project"] = dict()
        for col in self.cols:
            mongo_query = col.get_mongo()
            for key, item in mongo_query.items():
                pipeline_dict[key].update(item)

        return [{key: item} for key, item in pipeline_dict.items() if item]


class MongoOrdering:
    """MongoDB Query Node for Ordering"""

    def __init__(self, query: Query):
        self.query = query
        self.order = query.order_by

    def get_mongo_order(self, attname_as_key=False):
        key = "attname" if attname_as_key else "column"
        mongo_order = {}
        meta = self.query.get_meta()
        fields = {field.name: field for field in self.query.model._meta.get_fields()}
        for field in self.order or []:
            if field.startswith("-"):
                ordering = -1
                field = field[1:]
            else:
                ordering = 1
            field = meta.pk.attname if field == "pk" else field
            mongo_order.update({getattr(fields[field], key): ordering})
        return mongo_order



from urllib.parse import unquote, urlparse


def _extract_username_password(url: str) -> tuple[str, str, str]:
    url_components = urlparse(url)
    username, password = url_components.username, url_components.password
    if username and password:
        url_components = url_components._replace(
            netloc=url_components.hostname,
        )

    return username, unquote(password) if password else None, url_components.geturl()


def sanitize_client_opts(client_opts):
    # Sanitize client options.
    if "host" in client_opts:
        username, password, host = _extract_username_password(client_opts["host"])
        client_opts["username"] = username
        client_opts["password"] = password
        client_opts["host"] = host
    return client_opts



threadsafety = 2


class DataError(Exception):
    pass


class OperationalError(Exception):
    pass


class IntegrityError(Exception):
    pass


class InternalError(Exception):
    pass


class ProgrammingError(Exception):
    pass


class NotSupportedError(Exception):
    pass


class DatabaseError(Exception):
    pass


class InterfaceError(Exception):
    pass


class Error(Exception):
    pass



from typing import Generic, Literal, TypeVar

from django.db import models

T = TypeVar("T")


class MongoQuerySet(Generic[T], models.QuerySet[T]):
    """QuerySet which uses MongoDB as backend"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prefer_search = False
        self._aggregation_stages = []

    def prefer_search(self, prefer_search=True):
        obj = self._chain()
        obj._prefer_search = prefer_search
        obj.query.prefer_search = prefer_search
        return obj

    def add_aggregation_stage(
        self,
        stage: dict,
        position: Literal["prepend", "pre-sort", "append"] = "prepend",
    ):
        obj = self._chain()
        obj._aggregation_stages.append((position, stage))
        obj.query.aggregation_stages = obj._aggregation_stages
        return obj

    def _chain(self):
        """
        Add the _prefer_search hint to the chained query
        """
        obj = super()._chain()
        if obj._prefer_search:
            obj.query.prefer_search = obj._prefer_search
        if obj._aggregation_stages:
            obj.query.aggregation_stages = obj._aggregation_stages
        return obj

    def _clone(self):
        obj = super()._clone()
        obj._prefer_search = self._prefer_search
        obj._aggregation_stages = self._aggregation_stages
        return obj


class MongoManager(Generic[T], models.Manager[T]):
    """Manager which uses MongoDB as backend"""

    def get_queryset(self) -> MongoQuerySet[T]:
        return MongoQuerySet(self.model, using=self._db)

    def prefer_search(self, require_search=True) -> MongoQuerySet[T]:
        return self.get_queryset().prefer_search(require_search)



from django.db.backends.base.base import BaseDatabaseWrapper
from pymongo import MongoClient

import django_mongodb.database as Database
from django_mongodb.client import DatabaseClient
from django_mongodb.creation import DatabaseCreation
from django_mongodb.cursor import Cursor
from django_mongodb.features import DatabaseFeatures
from django_mongodb.introspection import DatabaseIntrospection
from django_mongodb.operations import DatabaseOperations
from django_mongodb.schema import DatabaseSchemaEditor


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = "django_mongodb"
    data_types = {
        "JSONField": "json",
        "ObjectIdField": "objectid",
        "ObjectIdAutoField": "objectid",
        "AutoField": "integer",
        "BigAutoField": "bigint",
        "BinaryField": "binary",
        "BooleanField": "boolean",
        "CharField": "string",
        "DateField": "date",
        "DateTimeField": "timestamp with time zone",
        "DecimalField": "decimal128",
        "DurationField": "interval",
        "FileField": "varchar(%(max_length)s)",
        "FilePathField": "varchar(%(max_length)s)",
        "FloatField": "double precision",
        "IntegerField": "integer",
        "BigIntegerField": "bigint",
        "IPAddressField": "inet",
        "GenericIPAddressField": "inet",
        "OneToOneField": "integer",
        "PositiveBigIntegerField": "bigint",
        "PositiveIntegerField": "integer",
        "PositiveSmallIntegerField": "smallint",
        "SlugField": "varchar(%(max_length)s)",
        "SmallAutoField": "smallint",
        "SmallIntegerField": "smallint",
        "TextField": "text",
        "TimeField": "time",
        "UUIDField": "uuid",
    }
    data_type_check_constraints = {
        "PositiveBigIntegerField": '"%(column)s" >= 0',
        "PositiveIntegerField": '"%(column)s" >= 0',
        "PositiveSmallIntegerField": '"%(column)s" >= 0',
    }
    data_types_suffix = {
        "AutoField": "GENERATED BY DEFAULT AS IDENTITY",
        "BigAutoField": "GENERATED BY DEFAULT AS IDENTITY",
        "SmallAutoField": "GENERATED BY DEFAULT AS IDENTITY",
    }
    operators = {
        "exact": "= %s",
        "iexact": "= UPPER(%s)",
        "contains": "LIKE %s",
        "icontains": "LIKE UPPER(%s)",
        "regex": "~ %s",
        "iregex": "~* %s",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
        "startswith": "LIKE %s",
        "endswith": "LIKE %s",
        "istartswith": "LIKE UPPER(%s)",
        "iendswith": "LIKE UPPER(%s)",
    }

    SchemaEditorClass = DatabaseSchemaEditor
    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations

    Database = Database

    def __init__(self, settings, *args, **kwargs):
        super().__init__(settings, *args, **kwargs)
        self.settings = settings
        self.mongo_client = MongoClient(**settings["CLIENT"])

    def get_connection_params(self):
        return self.settings_dict

    def get_new_connection(self, conn_params):
        name = conn_params.get("NAME") or "test"
        return self.mongo_client[name]

    def get_database_version(self):
        return self.connection.server_info()["version"]

    def create_cursor(self, name=None):
        return Cursor(self.mongo_client, self.connection)

    def _close(self):
        # if self.mongo_client is not None:
        #     with self.wrap_database_errors:
        #         self.mongo_client.close()
        pass

    def _set_autocommit(self, autocommit):
        pass

    def rollback(self):
        pass

    def commit(self):
        pass



import json
import logging

from pymongo import MongoClient
from pymongo.cursor import Cursor as MongoCursor
from pymongo.results import (
    DeleteResult,
    InsertManyResult,
    InsertOneResult,
    UpdateResult,
)

from django_mongodb.database import InterfaceError, NotSupportedError

logger = logging.getLogger(__name__)


class Cursor:
    def __init__(self, mongo_client: MongoClient, connection):
        self.mongo_client = mongo_client
        self.connection = connection
        self.result: MongoCursor | InsertManyResult | DeleteResult | None = None
        self.batch_size = None
        self.session = mongo_client.start_session()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if self.session:
            self.session.end_session()

    def close(self):
        if self.result is not None and hasattr(self.result, "close"):
            self.result.close()
        self.result = None

    @property
    def collections(self):
        return self.connection

    def __getattr__(self, name):
        try:
            return getattr(self.mongo_client, name)
        except AttributeError:
            pass

        try:
            return getattr(self.connection, name)
        except AttributeError:
            InterfaceError(f"Unsupported operation {name}")

    @property
    def rowcount(self):
        if self.cursor is None:
            raise RuntimeError
        if isinstance(self.result, InsertManyResult):
            return len(self.result.inserted_ids)
        if isinstance(self.result, DeleteResult):
            return self.result.deleted_count
        if isinstance(self.result, MongoCursor):
            if self.result.alive:
                return -1
            return self.result.retrieved
        if isinstance(self.result, UpdateResult):
            return self.result.matched_count  # update might be a no-op in case there are no changes
        raise NotSupportedError

    @property
    def lastrowid(self):
        if self.cursor is None:
            raise RuntimeError
        if isinstance(self.result, InsertOneResult):
            return self.result.inserted_id
        raise NotSupportedError

    def execute(self, command, params=None):
        logger.debug(json.dumps(command, default=str))
        match command:
            case {"op": "aggregate"}:
                self.result = self.connection[command["collection"]].aggregate(
                    command["pipeline"], session=self.session
                )
            case {"op": "insert_one"}:
                self.result = self.connection[command["collection"]].insert_one(
                    command["document"], session=self.session
                )
            case {"op": "update_many"}:
                self.result = self.connection[command["collection"]].update_many(
                    command["filter"], command["update"], session=self.session
                )
            case {"op": "bulk_write"}:
                self.result = self.connection[command["collection"]].bulk_write(
                    command["requests"], session=self.session
                )
            case {"op": "delete_many"}:
                self.result = self.connection[command["collection"]].delete_many(
                    command["filter"], session=self.session
                )
            case _:
                raise NotSupportedError

    def fetchmany(self, size=1):
        rows = []
        if self.batch_size != size:
            self.batch_size = size
            self.result.batch_size(size)
        for _ in range(size):
            try:
                rows.append(self.result.next())
            except StopIteration:
                return rows
        return rows

    def fetchone(self):
        try:
            return next(self.result)
        except StopIteration:
            return None

    def fetchall(self):
        return NotSupportedError

    def commit(self):
        pass



import bson
from django.db import models
from django.db.models.fields import AutoField, AutoFieldMeta
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _


class ObjectIdFieldMixin:
    description = "MongoDB ObjectIdField"
    default_error_messages = {
        "invalid": _("“%(value)s” value must be an ObjectId."),
    }

    @cached_property
    def validators(self):
        return self._validators

    def db_type(self, connection):
        return "ObjectId" if "ObjectIdField" in connection.data_types else "CHAR(24)"

    def to_python(self, value):
        if value is None or isinstance(value, bson.ObjectId):
            return value
        else:
            return bson.ObjectId(value)

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def get_prep_value(self, value):
        if value is None:
            return None
        if isinstance(value, str):
            return bson.ObjectId(value)
        return bson.ObjectId(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        return (
            self.get_prep_value(value) if "ObjectIdField" in connection.data_types else str(value)
        )


class ObjectIdField(ObjectIdFieldMixin, models.CharField):
    pass


class ObjectIdAutoField(ObjectIdFieldMixin, AutoField, metaclass=AutoFieldMeta):
    description = "MongoDB ObjectIdAutoField"

    def __init__(self, *args, **kwargs):
        if "db_column" not in kwargs:
            kwargs["db_column"] = "_id"
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "ObjectIdAutoField"

    def rel_db_type(self, connection):
        return ObjectIdField().db_type(connection=connection)
